@@ -15,6 +15,7 @@ typedef	struct gamelogic gamelogic;
 struct gamelogic {
     int		ruleset;		  /* the ruleset */
     gamestate  *state;			  /* ptr to the current game state */
+    int		localstateinfosize;	  /* how big to make localstateinfo */
     int	      (*initgame)(gamelogic*);	  /* prepare to play a game */
     int	      (*advancegame)(gamelogic*); /* advance the game one tick */
     int	      (*endgame)(gamelogic*);	  /* clean up after the game is done */
@@ -26,10 +27,4 @@ struct gamelogic {
 extern gamelogic *lynxlogicstartup(void);
 extern gamelogic *mslogicstartup(void);
 
-/* The high simluation fidelity flag: if true, the simulation should
- * forgo "standard play" in favor of being as true as possible to the
- * original source material.
- */
-extern int	pedanticmode;
-
 #endif
