@@ -25,8 +25,7 @@
  * HEADER
  *  0-3   signature bytes (35 33 9B 99)
  *   4    ruleset (1=Lynx, 2=MS)
- *   5    other options (currently always zero)
- *   6    other options (currently always zero)
+ *  5-6   most recently visited level number
  *   7    count of bytes in remainder of header (currently always zero)
  *
  * After the header are level solutions, usually but not necessarily
@@ -119,6 +118,10 @@
  */
 #define	CSSIG		0x999B3335UL
 
+/* The three different modes that solutions files are opened with.
+ */
+enum { F_READ, F_WRITE, F_MODIFY };
+
 /* Translate move directions between three-bit and four-bit
  * representations.
  *
@@ -146,13 +149,25 @@ static int const idxdir8[8] = {
 #define	dirtoindex(dir)		(diridx8[dir])
 #define	indextodir(dir)		(idxdir8[dir])
 
-/* TRUE if file modification is prohibited.
- */
-int		readonly = FALSE;
-
 /* The path of the directory containing the user's solution files.
  */
-char	       *savedir = NULL;
+static char const      *savedir = NULL;
+
+/* TRUE if file modification is prohibited.
+ */
+static int		readonly = FALSE;
+
+/* Getting and setting the save directory.
+ */
+char const *getsavedir(void)		{ return savedir; }
+void setsavedir(char const *dir)	{ savedir = dir; }
+
+/* Put the system in read-only mode.
+ */
+void setreadonly(void)
+{
+    readonly = TRUE;
+}
 
 /*
  * Functions for manipulating move lists.
@@ -362,7 +377,7 @@ int contractsolution(solutioninfo const *solution, gamesetup *game)
 {
     action const       *move;
     unsigned char      *data;
-    int			size, est, delta, when, i;
+    int			size, delta, when, i;
 
     free(game->solutiondata);
     game->solutionsize = 0;
@@ -382,7 +397,6 @@ int contractsolution(solutioninfo const *solution, gamesetup *game)
 		     " out of memory", game->number);
 	return FALSE;
     }
-    est = size;
 
     data[0] = game->number & 0xFF;
     data[1] = (game->number >> 8) & 0xFF;
@@ -544,14 +558,14 @@ static int writesolution(fileinfo *file, gamesetup const *game)
 
 /* Locate the solution file for the given data file and open it.
  */
-static int opensolutionfile(fileinfo *file, char const *datname, int writable)
+static int opensolutionfile(fileinfo *file, char const *datname, int mode)
 {
     static int	savedirchecked = FALSE;
     char       *buf = NULL;
     char const *filename;
     int		n;
 
-    if (writable && readonly)
+    if (mode != F_READ && readonly)
 	return FALSE;
 
     if (file->name) {
@@ -568,19 +582,19 @@ static int opensolutionfile(fileinfo *file, char const *datname, int writable)
 	filename = buf;
     }
 
-    if (writable) {
+    if (mode != F_WRITE) {
 	if (!savedirchecked && savedir && *savedir && !haspathname(filename)) {
 	    savedirchecked = TRUE;
 	    if (!finddir(savedir)) {
-		*savedir = '\0';
+		setsavedir("");
 		fileerr(file, "can't access directory");
 	    }
 	}
     }
 
     n = openfileindir(file, savedir, filename,
-		      writable ? "wb" : "rb",
-		      writable ? "can't access file" : NULL);
+		      mode == F_WRITE ? "wb" : mode == F_MODIFY ? "r+b" : "rb",
+		      mode == F_WRITE ? "can't access file" : NULL);
     if (buf)
 	free(buf);
     return n;
@@ -597,16 +611,19 @@ int readsolutions(gameseries *series)
 	series->savefile.name = series->savefilename;
     if ((!series->savefile.name && (series->gsflags & GSF_NODEFAULTSAVE))
 		|| !opensolutionfile(&series->savefile,
-				     series->filebase, FALSE)) {
-	series->solheaderflags = 0;
+				     series->filebase, F_READ)) {
+	series->currentlevel = 0;
 	series->solheadersize = 0;
 	return TRUE;
     }
 
     if (!readsolutionheader(&series->savefile, series->ruleset,
-			    &series->solheaderflags,
-			    &series->solheadersize, series->solheader))
+			    &series->currentlevel,
+			    &series->solheadersize, series->solheader)) {
+	series->currentlevel = 0;
+	series->solheadersize = 0;
 	return FALSE;
+    }
 
     memset(&gametmp, 0, sizeof gametmp);
     for (;;) {
@@ -659,11 +676,11 @@ int savesolutions(gameseries *series)
 	series->savefile.name = series->savefilename;
     if (!series->savefile.name && (series->gsflags & GSF_NODEFAULTSAVE))
 	return TRUE;
-    if (!opensolutionfile(&series->savefile, series->filebase, TRUE))
+    if (!opensolutionfile(&series->savefile, series->filebase, F_WRITE))
 	return FALSE;
 
     if (!writesolutionheader(&series->savefile, series->ruleset,
-			     series->solheaderflags,
+			     series->currentlevel,
 			     series->solheadersize, series->solheader))
 	return fileerr(&series->savefile,
 		       "saved-game file has become corrupted!");
@@ -680,6 +697,27 @@ int savesolutions(gameseries *series)
     return TRUE;
 }
 
+/* Write out just the current level number to the existing solution
+ * file.
+ */
+int savesolutionlevel(gameseries *series)
+{
+    int f;
+
+    if (readonly)
+	return FALSE;
+    if (!series->savefile.name)
+	series->savefile.name = series->savefilename;
+    if (!series->savefile.name && (series->gsflags & GSF_NODEFAULTSAVE))
+	return FALSE;
+    if (!opensolutionfile(&series->savefile, series->filebase, F_MODIFY))
+	return FALSE;
+    f = fileskip(&series->savefile, 5, NULL)
+	    && filewriteint16(&series->savefile, series->currentlevel, NULL);
+    fileclose(&series->savefile, NULL);
+    return f;
+}
+
 /* Free all memory allocated for storing the game's solutions, and mark
  * the levels as being unsolved.
  */
@@ -696,7 +734,7 @@ void clearsolutions(gameseries *series)
 	game->solutiondata = NULL;
     }
     series->solheadersize = 0;
-    series->solheaderflags = 0;
+    series->currentlevel = 0;
     fileclose(&series->savefile, NULL);
     clearfileinfo(&series->savefile);
 }
@@ -819,7 +857,7 @@ int createsolutionfilelist(gameseries const *series, int morethanone,
 	memerrexit();
     table->rows = s.count + 1;
     table->cols = 2;
-    table->sep = 4;
+    table->sep = 1;
     table->collapse = 1;
     table->items[0] = "2-Select a solution file";
     offset = 0;
